@@ -7,7 +7,7 @@ coverage reports.  There's a third convenient decorator (`timecall`) that
 measures the duration of function execution without the extra profiling
 overhead.
 
-Usage example (Python 2.4 or newer)::
+Usage example::
 
     from profilehooks import profile, coverage
 
@@ -29,20 +29,6 @@ Or without imports, with some hack
 
     print(fn(42))
 
-
-Usage example (Python 2.3 or older)::
-
-    from profilehooks import profile, coverage
-
-    def fn(n):
-        if n < 2: return 1
-        else: return n * fn(n-1)
-
-    # Now wrap that function in a decorator
-    fn = profile(fn) # or coverage(fn)
-
-    print fn(42)
-
 Reports for all thusly decorated functions will be printed to sys.stdout
 on program termination.  You can alternatively request for immediate
 reports for each call by passing immediate=True to the profile decorator.
@@ -54,7 +40,7 @@ instead of a detailed (but costly) profile.
 Caveats
 
   A thread on python-dev convinced me that hotshot produces bogus numbers.
-  See http://mail.python.org/pipermail/python-dev/2005-November/058264.html
+  See https://mail.python.org/pipermail/python-dev/2005-November/058264.html
 
   I don't know what will happen if a decorated function will try to call
   another decorated function.  All decorators probably need to explicitly
@@ -74,7 +60,7 @@ Caveats
   executed.  For this reason coverage analysis now uses trace.py which is
   slower, but more accurate.
 
-Copyright (c) 2004--2014 Marius Gedminas <marius@pov.lt>
+Copyright (c) 2004--2020 Marius Gedminas <marius@gedmin.as>
 Copyright (c) 2007 Hanno Schlichting
 Copyright (c) 2008 Florian Schulze
 
@@ -100,23 +86,29 @@ Released under the MIT licence since December 2006:
 
 (Previously it was distributed under the GNU General Public Licence.)
 """
+from __future__ import print_function
 
 __author__ = "Marius Gedminas <marius@gedmin.as>"
-__copyright__ = "Copyright 2004-2014 Marius Gedminas"
+__copyright__ = "Copyright 2004-2020 Marius Gedminas and contributors"
 __license__ = "MIT"
-__version__ = "1.7.1"
-__date__ = "2014-12-02"
-
+__version__ = '1.11.2'
+__date__ = "2020-03-03"
 
 import atexit
+import functools
 import inspect
-import sys
+import logging
+import os
 import re
+import sys
 
 # For profiling
 from profile import Profile
 import pstats
 
+# For timecall
+import timeit
+
 # For hotshot profiling (inaccurate!)
 try:
     import hotshot
@@ -126,6 +118,9 @@ except ImportError:
 
 # For trace.py coverage
 import trace
+import dis
+import token
+import tokenize
 
 # For hotshot coverage (inaccurate!; uses undocumented APIs; might break)
 if hotshot is not None:
@@ -138,9 +133,6 @@ try:
 except ImportError:
     cProfile = None
 
-# For timecall
-import time
-
 
 # registry of available profilers
 AVAILABLE_PROFILERS = {}
@@ -148,16 +140,40 @@ AVAILABLE_PROFILERS = {}
 __all__ = ['coverage', 'coverage_with_hotshot', 'profile', 'timecall']
 
 
+# Use tokenize.open() on Python >= 3.2, fall back to open() on Python 2
+tokenize_open = getattr(tokenize, 'open', open)
+
+
+def _unwrap(fn):
+    # inspect.unwrap() doesn't exist on Python 2
+    if not hasattr(fn, '__wrapped__'):
+        return fn
+    else:
+        # intentionally using recursion here instead of a while loop to
+        # make cycles fail with a recursion error instead of looping forever.
+        return _unwrap(fn.__wrapped__)
+
+
+def _identify(fn):
+    fn = _unwrap(fn)
+    funcname = fn.__name__
+    filename = fn.__code__.co_filename
+    lineno = fn.__code__.co_firstlineno
+    return (funcname, filename, lineno)
+
+
 def profile(fn=None, skip=0, filename=None, immediate=False, dirs=False,
             sort=None, entries=40,
-            profiler=('cProfile', 'profile', 'hotshot')):
+            profiler=('cProfile', 'profile', 'hotshot'),
+            stdout=True):
     """Mark `fn` for profiling.
 
     If `skip` is > 0, first `skip` calls to `fn` will not be profiled.
 
     If `immediate` is False, profiling results will be printed to
     sys.stdout on program termination.  Otherwise results will be printed
-    after each call.
+    after each call.  (If you don't want this, set stdout=False and specify a
+    `filename` to store profile data.)
 
     If `dirs` is False only the name of the file will be printed.
     Otherwise the full path is used.
@@ -183,7 +199,8 @@ def profile(fn=None, skip=0, filename=None, immediate=False, dirs=False,
     'profile', 'hotshot').
 
     If `filename` is specified, the profile stats will be stored in the
-    named file.  You can load them pstats.Stats(filename).
+    named file.  You can load them with pstats.Stats(filename) or use a
+    visualization tool like RunSnakeRun.
 
     Usage::
 
@@ -210,7 +227,7 @@ def profile(fn=None, skip=0, filename=None, immediate=False, dirs=False,
             return profile(fn, skip=skip, filename=filename,
                            immediate=immediate, dirs=dirs,
                            sort=sort, entries=entries,
-                           profiler=profiler)
+                           profiler=profiler, stdout=stdout)
         return decorator
     # @profile syntax -- we are a decorator.
     if isinstance(profiler, str):
@@ -221,19 +238,16 @@ def profile(fn=None, skip=0, filename=None, immediate=False, dirs=False,
             break
     else:
         raise ValueError('only these profilers are available: %s'
-                             % ', '.join(sorted(AVAILABLE_PROFILERS)))
+                         % ', '.join(sorted(AVAILABLE_PROFILERS)))
     fp = profiler_class(fn, skip=skip, filename=filename,
                         immediate=immediate, dirs=dirs,
-                        sort=sort, entries=entries)
+                        sort=sort, entries=entries, stdout=stdout)
     # We cannot return fp or fp.__call__ directly as that would break method
     # definitions, instead we need to return a plain function.
 
+    @functools.wraps(fn)
     def new_fn(*args, **kw):
         return fp(*args, **kw)
-    new_fn.__doc__ = fn.__doc__
-    new_fn.__name__ = fn.__name__
-    new_fn.__dict__ = fn.__dict__
-    new_fn.__module__ = fn.__module__
     return new_fn
 
 
@@ -260,12 +274,9 @@ def coverage(fn):
     # We cannot return fp or fp.__call__ directly as that would break method
     # definitions, instead we need to return a plain function.
 
+    @functools.wraps(fn)
     def new_fn(*args, **kw):
         return fp(*args, **kw)
-    new_fn.__doc__ = fn.__doc__
-    new_fn.__name__ = fn.__name__
-    new_fn.__dict__ = fn.__dict__
-    new_fn.__module__ = fn.__module__
     return new_fn
 
 
@@ -282,12 +293,9 @@ def coverage_with_hotshot(fn):
     # We cannot return fp or fp.__call__ directly as that would break method
     # definitions, instead we need to return a plain function.
 
+    @functools.wraps(fn)
     def new_fn(*args, **kw):
         return fp(*args, **kw)
-    new_fn.__doc__ = fn.__doc__
-    new_fn.__name__ = fn.__name__
-    new_fn.__dict__ = fn.__dict__
-    new_fn.__module__ = fn.__module__
     return new_fn
 
 
@@ -300,7 +308,7 @@ class FuncProfile(object):
     Profile = Profile
 
     def __init__(self, fn, skip=0, filename=None, immediate=False, dirs=False,
-                 sort=None, entries=40):
+                 sort=None, entries=40, stdout=True):
         """Creates a profiler for a function.
 
         Every profiler has its own log file (the name of which is derived
@@ -312,14 +320,20 @@ class FuncProfile(object):
         self.fn = fn
         self.skip = skip
         self.filename = filename
-        self.immediate = immediate
+        self._immediate = immediate
+        self.stdout = stdout
         self.dirs = dirs
         self.sort = sort or ('cumulative', 'time', 'calls')
         if isinstance(self.sort, str):
             self.sort = (self.sort, )
         self.entries = entries
         self.reset_stats()
-        atexit.register(self.atexit)
+        if not self.immediate:
+            atexit.register(self.atexit)
+
+    @property
+    def immediate(self):
+        return self._immediate
 
     def __call__(self, *args, **kw):
         """Profile a singe call to the function."""
@@ -346,25 +360,24 @@ class FuncProfile(object):
 
     def print_stats(self):
         """Print profile information to sys.stdout."""
-        funcname = self.fn.__name__
-        filename = self.fn.__code__.co_filename
-        lineno = self.fn.__code__.co_firstlineno
-        print("")
-        print("*** PROFILER RESULTS ***")
-        print("%s (%s:%s)" % (funcname, filename, lineno))
-        if self.skipped:
-            skipped = " (%d calls not profiled)" % self.skipped
-        else:
-            skipped = ""
-        print("function called %d times%s" % (self.ncalls, skipped))
-        print("")
         stats = self.stats
         if self.filename:
             stats.dump_stats(self.filename)
-        if not self.dirs:
-            stats.strip_dirs()
-        stats.sort_stats(*self.sort)
-        stats.print_stats(self.entries)
+        if self.stdout:
+            funcname, filename, lineno = _identify(self.fn)
+            print("")
+            print("*** PROFILER RESULTS ***")
+            print("%s (%s:%s)" % (funcname, filename, lineno))
+            if self.skipped:
+                skipped = " (%d calls not profiled)" % self.skipped
+            else:
+                skipped = ""
+            print("function called %d times%s" % (self.ncalls, skipped))
+            print("")
+            if not self.dirs:
+                stats.strip_dirs()
+            stats.sort_stats(*self.sort)
+            stats.print_stats(self.entries)
 
     def reset_stats(self):
         """Reset accumulated profiler statistics."""
@@ -378,8 +391,7 @@ class FuncProfile(object):
 
         This function is registered as an atexit hook.
         """
-        if not self.immediate:
-            self.print_stats()
+        self.print_stats()
 
 
 AVAILABLE_PROFILERS['profile'] = FuncProfile
@@ -404,7 +416,7 @@ if hotshot is not None:
         in_profiler = False
 
         def __init__(self, fn, skip=0, filename=None, immediate=False,
-                     dirs=False, sort=None, entries=40):
+                     dirs=False, sort=None, entries=40, stdout=True):
             """Creates a profiler for a function.
 
             Every profiler has its own log file (the name of which is derived
@@ -419,10 +431,10 @@ if hotshot is not None:
             if filename:
                 self.logfilename = filename + ".raw"
             else:
-                self.logfilename = fn.__name__ + ".prof"
+                self.logfilename = "%s.%d.prof" % (fn.__name__, os.getpid())
             super(HotShotFuncProfile, self).__init__(
                 fn, skip=skip, filename=filename, immediate=immediate,
-                dirs=dirs, sort=sort, entries=entries)
+                dirs=dirs, sort=sort, entries=entries, stdout=stdout)
 
         def __call__(self, *args, **kw):
             """Profile a singe call to the function."""
@@ -482,7 +494,7 @@ if hotshot is not None:
             current working directory.
             """
             self.fn = fn
-            self.logfilename = fn.__name__ + ".cprof"
+            self.logfilename = "%s.%d.cprof" % (fn.__name__, os.getpid())
             self.profiler = _hotshot.coverage(self.logfilename)
             self.ncalls = 0
             atexit.register(self.atexit)
@@ -493,7 +505,7 @@ if hotshot is not None:
             old_trace = sys.gettrace()
             try:
                 return self.profiler.runcall(self.fn, args, kw)
-            finally: # pragma: nocover
+            finally:  # pragma: nocover
                 sys.settrace(old_trace)
 
         def atexit(self):
@@ -502,9 +514,7 @@ if hotshot is not None:
             This function is registered as an atexit hook.
             """
             self.profiler.close()
-            funcname = self.fn.__name__
-            filename = self.fn.__code__.co_filename
-            lineno = self.fn.__code__.co_firstlineno
+            funcname, filename, lineno = _identify(self.fn)
             print("")
             print("*** COVERAGE RESULTS ***")
             print("%s (%s:%s)" % (funcname, filename, lineno))
@@ -518,10 +528,11 @@ if hotshot is not None:
                 if what == hotshot.log.LINE:
                     fs.mark(lineno)
                 if what == hotshot.log.ENTER:
-                    # hotshot gives us the line number of the function definition
-                    # and never gives us a LINE event for the first statement in
-                    # a function, so if we didn't perform this mapping, the first
-                    # statement would be marked as never executed
+                    # hotshot gives us the line number of the function
+                    # definition and never gives us a LINE event for the first
+                    # statement in a function, so if we didn't perform this
+                    # mapping, the first statement would be marked as never
+                    # executed
                     if lineno == fs.firstlineno:
                         lineno = fs.firstcodelineno
                     fs.mark(lineno)
@@ -559,20 +570,20 @@ class TraceFuncCoverage:
         current working directory.
         """
         self.fn = fn
-        self.logfilename = fn.__name__ + ".cprof"
+        self.logfilename = "%s.%d.cprof" % (fn.__name__, os.getpid())
         self.ncalls = 0
         atexit.register(self.atexit)
 
     def __call__(self, *args, **kw):
         """Profile a singe call to the function."""
         self.ncalls += 1
-        if TraceFuncCoverage.tracing: # pragma: nocover
+        if TraceFuncCoverage.tracing:  # pragma: nocover
             return self.fn(*args, **kw)
         old_trace = sys.gettrace()
         try:
             TraceFuncCoverage.tracing = True
             return self.tracer.runfunc(self.fn, *args, **kw)
-        finally: # pragma: nocover
+        finally:  # pragma: nocover
             sys.settrace(old_trace)
             TraceFuncCoverage.tracing = False
 
@@ -581,9 +592,7 @@ class TraceFuncCoverage:
 
         This function is registered as an atexit hook.
         """
-        funcname = self.fn.__name__
-        filename = self.fn.__code__.co_filename
-        lineno = self.fn.__code__.co_firstlineno
+        funcname, filename, lineno = _identify(self.fn)
         print("")
         print("*** COVERAGE RESULTS ***")
         print("%s (%s:%s)" % (funcname, filename, lineno))
@@ -622,8 +631,12 @@ class FuncSource:
         """Mark all executable source lines in fn as executed 0 times."""
         if self.filename is None:
             return
-        strs = trace.find_strings(self.filename)
-        lines = trace.find_lines_from_code(self.fn.__code__, strs)
+        strs = self._find_docstrings(self.filename)
+        lines = {
+            ln
+            for off, ln in dis.findlinestarts(_unwrap(self.fn).__code__)
+            if ln not in strs
+        }
         for lineno in lines:
             self.sourcelines.setdefault(lineno, 0)
         if lines:
@@ -632,6 +645,19 @@ class FuncSource:
             # This branch cannot be reached, I'm just being paranoid.
             self.firstcodelineno = self.firstlineno
 
+    def _find_docstrings(self, filename):
+        # A replacement for trace.find_strings() which was deprecated in
+        # Python 3.2 and removed in 3.6.
+        strs = set()
+        prev = token.INDENT  # so module docstring is detected as docstring
+        with tokenize_open(filename) as f:
+            tokens = tokenize.generate_tokens(f.readline)
+            for ttype, tstr, start, end, line in tokens:
+                if ttype == token.STRING and prev == token.INDENT:
+                    strs.update(range(start[0], end[0] + 1))
+                prev = ttype
+        return strs
+
     def mark(self, lineno, count=1):
         """Mark a given source line as executed count times.
 
@@ -675,7 +701,10 @@ class FuncSource:
         return ''.join(lines)
 
 
-def timecall(fn=None, immediate=True, timer=None):
+def timecall(
+        fn=None, immediate=True, timer=None,
+        log_name=None, log_level=logging.DEBUG,
+):
     """Wrap `fn` and print its execution time.
 
     Example::
@@ -687,39 +716,56 @@ def timecall(fn=None, immediate=True, timer=None):
         somefunc(2, 3)
 
     will print the time taken by somefunc on every call.  If you want just
-    a summary at program termination, use
+    a summary at program termination, use ::
 
         @timecall(immediate=False)
 
-    You can also choose a timing method other than the default ``time.time()``,
-    e.g.:
+    You can also choose a timing method other than the default
+    ``timeit.default_timer()``, e.g.::
 
         @timecall(timer=time.clock)
 
+    You can also log the output to a logger by specifying the name and level
+    of the logger to use, eg:
+
+        @timecall(immediate=True,
+                  log_name='profile_log',
+                  log_level=logging.DEBUG)
+
     """
     if fn is None:  # @timecall() syntax -- we are a decorator maker
         def decorator(fn):
-            return timecall(fn, immediate=immediate, timer=timer)
+            return timecall(
+                fn, immediate=immediate, timer=timer,
+                log_name=log_name, log_level=log_level,
+            )
         return decorator
     # @timecall syntax -- we are a decorator.
     if timer is None:
-        timer = time.time
-    fp = FuncTimer(fn, immediate=immediate, timer=timer)
+        timer = timeit.default_timer
+    fp = FuncTimer(
+        fn, immediate=immediate, timer=timer,
+        log_name=log_name, log_level=log_level,
+    )
     # We cannot return fp or fp.__call__ directly as that would break method
     # definitions, instead we need to return a plain function.
 
+    @functools.wraps(fn)
     def new_fn(*args, **kw):
         return fp(*args, **kw)
-    new_fn.__doc__ = fn.__doc__
-    new_fn.__name__ = fn.__name__
-    new_fn.__dict__ = fn.__dict__
-    new_fn.__module__ = fn.__module__
     return new_fn
 
 
 class FuncTimer(object):
 
-    def __init__(self, fn, immediate, timer):
+    def __init__(
+            self, fn, immediate, timer,
+            log_name=None, log_level=logging.DEBUG,
+    ):
+        self.logger = None
+        if log_name:
+            self.logger = logging.getLogger(log_name)
+        self.log_level = log_level
         self.fn = fn
         self.ncalls = 0
         self.totaltime = 0
@@ -733,32 +779,36 @@ class FuncTimer(object):
         fn = self.fn
         timer = self.timer
         self.ncalls += 1
+        start = timer()
         try:
-            start = timer()
             return fn(*args, **kw)
         finally:
             duration = timer() - start
             self.totaltime += duration
             if self.immediate:
-                funcname = fn.__name__
-                filename = fn.__code__.co_filename
-                lineno = fn.__code__.co_firstlineno
-                sys.stderr.write("\n  %s (%s:%s):\n    %.3f seconds\n\n" % (
-                    funcname, filename, lineno, duration
-                ))
-                sys.stderr.flush()
+                funcname, filename, lineno = _identify(fn)
+                message = "%s (%s:%s):\n    %.3f seconds\n\n" % (
+                    funcname, filename, lineno, duration,
+                )
+                if self.logger:
+                    self.logger.log(self.log_level, message)
+                else:
+                    sys.stderr.write("\n  " + message)
+                    sys.stderr.flush()
 
     def atexit(self):
         if not self.ncalls:
             return
-        funcname = self.fn.__name__
-        filename = self.fn.__code__.co_filename
-        lineno = self.fn.__code__.co_firstlineno
-        print("\n  %s (%s:%s):\n"
-              "    %d calls, %.3f seconds (%.3f seconds per call)\n" % (
-                  funcname, filename, lineno, self.ncalls,
-                  self.totaltime, self.totaltime / self.ncalls)
-              )
+        funcname, filename, lineno = _identify(self.fn)
+        message = "\n  %s (%s:%s):\n"\
+            "    %d calls, %.3f seconds (%.3f seconds per call)\n" % (
+                funcname, filename, lineno, self.ncalls,
+                self.totaltime, self.totaltime / self.ncalls)
+        if self.logger:
+            self.logger.log(self.log_level, message)
+        else:
+            print(message)
+
 
 if __name__ == '__main__':
 
