@@ -7,21 +7,27 @@ They are woefully incomplete.
 Run it with python setup.py test
 """
 
+import atexit
+import doctest
+import functools
+import inspect
+import logging
 import os
 import sys
-import doctest
-import unittest
-import linecache
-import inspect
-import atexit
 import textwrap
-import time
+import timeit
+import unittest
 
 try:
     from cStringIO import StringIO
 except ImportError:
     from io import StringIO
 
+try:
+    import pytest
+except ImportError:
+    pytest = None
+
 import profilehooks
 
 
@@ -98,13 +104,14 @@ class TestCoverage(TestCase):
         sample_fn = self.decorator(self.sample_fn)
         sample_fn(1, 1, 1)
         sample_fn(1, 2, 3)
+        filename = self.sample_fn.__code__.co_filename
         linenumber = self.sample_fn.__code__.co_firstlineno
         run_exitfuncs()
         self.assertMultiLineEqual(
             sys.stdout.getvalue(),
             '\n' + textwrap.dedent("""\
             *** COVERAGE RESULTS ***
-            sample_fn (test_profilehooks.py:{0})
+            sample_fn ({filename}:{lineno})
             function called 2 times
 
                        def sample_fn(self, x, y, z):
@@ -116,18 +123,19 @@ class TestCoverage(TestCase):
                 1:             return "%s %s %s" % (x, y, z)
 
             1 lines were not executed.
-            """.format(linenumber)))
+            """.format(filename=filename, lineno=linenumber)))
 
     def test_coverage_again(self):
         sample_fn_2 = self.decorator(self.sample_fn_2)
         sample_fn_2()
+        filename = self.sample_fn_2.__code__.co_filename
         linenumber = self.sample_fn_2.__code__.co_firstlineno
         run_exitfuncs()
         self.assertMultiLineEqual(
             sys.stdout.getvalue(),
             '\n' + textwrap.dedent("""\
             *** COVERAGE RESULTS ***
-            sample_fn_2 (test_profilehooks.py:{0})
+            sample_fn_2 ({filename}:{lineno})
             function called 1 times
 
                        def sample_fn_2(self):
@@ -137,18 +145,21 @@ class TestCoverage(TestCase):
                 1:             x = 5
                 1:         del x
 
-            """.format(linenumber)))
+            """.format(filename=filename, lineno=linenumber)))
+
 
 if profilehooks.hotshot is not None:
+
     class TestCoverageWithHotShot(TestCoverage):
         decorator = staticmethod(profilehooks.coverage_with_hotshot)
 
         def tearDown(self):
             super(TestCoverageWithHotShot, self).tearDown()
-            if os.path.exists('sample_fn.cprof'):
-                os.unlink('sample_fn.cprof')
-            if os.path.exists('sample_fn_2.cprof'):
-                os.unlink('sample_fn_2.cprof')
+            for name in 'sample_fn', 'sample_fn_2':
+                try:
+                    os.unlink('%s.%d.cprof' % (name, os.getpid()))
+                except OSError:
+                    pass
 
 
 def doctest_coverage_when_source_is_not_available(self):
@@ -320,7 +331,7 @@ def doctest_profile_with_hotshot():
 
     Hotshot leaves temporary files behind
 
-        >>> os.unlink('fac.prof')
+        >>> os.unlink('fac.%d.prof' % os.getpid())
 
     """
 
@@ -437,6 +448,104 @@ def doctest_timecall_never_called():
     """
 
 
+if hasattr('functools', 'lru_cache'):
+    # Python 2.7 doesn't have functools.lru_cache
+
+    def doctest_timecall_on_lru_cache():
+        """Test for timecall.
+
+        This is a regression test for
+        https://github.com/mgedmin/profilehooks/issues/25.
+
+            >>> @profilehooks.timecall
+            ... @functools.lru_cache()
+            ... def sample_fn(x, y, z):
+            ...     print("%s %s %s" % (x, y, z))
+            ...     return x + y * z
+
+        You can call that function normally
+
+            >>> r = sample_fn(1, 2, z=3)
+            1 2 3
+            >>> r
+            7
+
+        Every call also prints to stderr
+
+            >>> print(sys.stderr.getvalue())
+            <BLANKLINE>
+              sample_fn (<doctest test_profilehooks.doctest_timecall[0]>:1):
+                0.000 seconds
+            <BLANKLINE>
+            <BLANKLINE>
+
+        """
+
+
+def doctest_timecall_to_log_immediate():
+    """Test for timecall.
+
+        >>> logger = logging.getLogger('test_profilehooks')
+        >>> logger.propagate = False
+        >>> logger.level = logging.DEBUG
+        >>> buffer = StringIO()
+        >>> handler = logging.StreamHandler(buffer)
+        >>> logger.addHandler(handler)
+
+        >>> @profilehooks.timecall(log_name='test_profilehooks')
+        ... def sample_fn(x, y, z):
+        ...     print("%s %s %s" % (x, y, z))
+        ...     return x + y * z
+
+        >>> sample_fn(3, 2, 1)
+        3 2 1
+        5
+
+        >>> print(buffer.getvalue().rstrip())
+        sample_fn (<doctest test_profilehooks.doctest_timecall_to_log_immediate[6]>:1):
+            0.000 seconds
+
+        >>> run_exitfuncs()
+
+        >>> del logger.handlers[:]
+
+    """
+
+
+def doctest_timecall_to_log_not_immediate():
+    """Test for timecall.
+
+        >>> logger = logging.getLogger('test_profilehooks')
+        >>> logger.propagate = False
+        >>> logger.level = logging.DEBUG
+        >>> buffer = StringIO()
+        >>> handler = logging.StreamHandler(buffer)
+        >>> logger.addHandler(handler)
+
+        >>> @profilehooks.timecall(log_name='test_profilehooks',
+        ...                        immediate=False)
+        ... def sample_fn(x, y, z):
+        ...     print("%s %s %s" % (x, y, z))
+        ...     return x + y * z
+
+        >>> sample_fn(3, 2, 1)
+        3 2 1
+        5
+
+        >>> print(buffer.getvalue().strip())
+        <BLANKLINE>
+
+        >>> run_exitfuncs()
+
+        >>> print(buffer.getvalue().strip())
+        sample_fn (<doctest test_profilehooks.doctest_timecall_to_log_not_immediate[6]>:1):
+            1 calls, 0.000 seconds (0.000 seconds per call)
+
+        >>> del logger.handlers[:]
+
+    """
+
+
 def doctest_dump():
     """Test that profiling can save the stats in a file.
 
@@ -477,15 +586,30 @@ def setUp(test):
     sys.stderr = stderr_wrapper
     atexit.register = _register_exitfunc
     del _exitfuncs[:]
-    test.real_time = time.time
-    time.time = lambda: 1411735756
+    test.real_time = timeit.default_timer
+    timeit.default_timer = lambda: 1411735756
 
 
 def tearDown(test):
     sys.stderr = test.real_stderr
     atexit.register = test.real_register
     del _exitfuncs[:]
-    time.time = test.real_time
+    timeit.default_timer = test.real_time
+
+
+if pytest is not None:
+
+    class Bag(object):
+        pass
+
+    @pytest.yield_fixture(autouse=True)
+    def setUpForPyTest(capsys):
+        # NB: we need capsys otherwise our custom sys.stderr wrapper that we set up
+        # in the setUp() call here will be replaced by pytest's own wrapper.
+        bag = Bag()
+        setUp(bag)
+        yield
+        tearDown(bag)
 
 
 def additional_tests():
@@ -494,13 +618,12 @@ def additional_tests():
     return unittest.TestSuite([
         unittest.defaultTestLoader.loadTestsFromName(__name__),
         doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
-                                optionflags=optionflags),
+                             optionflags=optionflags),
     ])
 
 
 if __name__ == '__main__':
-    # a bit pointless: __name__ is different and thus all tests will fail
+    # Haaack!  We overwrite: __name__ so that tests will not fail
     __name__ = 'test_profilehooks'
     sys.modules[__name__] = sys.modules['__main__']
     unittest.main(defaultTest='additional_tests')
-
